from tek_utils.sharehoster.torrent import torrent_cacher

from tek.util.decorator import lazy_property
from golgi.config import configurable


@configurable(torrent=['cacher'])
class TorrentProxy(object):

    def __init__(self, record):
        self.record = record
        self.link = record.url

    @lazy_property
    def cacher(self):
        return torrent_cacher(self.link)

    @property
    def domain(self):
        self._cacher

    @property
    def cachable(self):
        return not (self.caching or self.cached)

    @property
    def caching(self):
        return self.valid and self.cacher.caching

    @property
    def cached(self):
        return self.valid and self.cacher.downloaded

    def request(self):
        if self.valid:
            self.cacher.request()

    @property
    def valid(self):
        return self.cacher is not None

    @property
    def download_url(self):
        if self.valid:
            return self.cacher.download_url

__all__ = ['TorrentProxy']



import re
from datetime import datetime
import calendar

from tek.errors import ParseError


def domain(url):
    """ Extract the domain name out of url.
    """
    if isinstance(url, str):
        domain_match = re.match(r'https?://(?:www\.)?([^/]+)\.[^/]+', url)
        return domain_match.group(1) if domain_match else ''
    else:
        raise ParseError('Invalid input for domain(): {}'.format(url))


def unix_to_datetime(stamp):
    return datetime.utcfromtimestamp(stamp)


def datetime_to_unix(_date):
    return calendar.timegm(_date.utctimetuple())


def now_unix():
    return datetime_to_unix(datetime.now())

__all__ = ('domain', 'unix_to_datetime', 'datetime_to_unix')



import socket
from multiprocessing import Process, Value, Array  # type: ignore
from datetime import datetime, timedelta

import requests

from golgi.config import configurable
from tek.errors import NotEnoughDiskSpace
from tek.tools import sizeof_fmt

from tek_utils.sharehoster import DownloaderFactory
from tek_utils.sharehoster.errors import ShareHosterError

from amino import List, Empty, Just, Path

from series.get.handler import ReleaseHandler, R, L
from series.condition import DynOr, HandlerCondition
from series.get.model.link import Link, Torrent
from series.get.model.release import ReleaseMonitor
from series.logging import Logging
from series.util import now_unix


class ValidLink(HandlerCondition):

    def __init__(self, link: Link) -> None:
        self.link = link

    @property
    def cond(self):
        if isinstance(self.link, Torrent):
            return self._cond_torrent
        else:
            return self._cond_http

    @property
    def _cond_torrent(self):
        return (~(L('url', None) | L('dead') | L('download_url', None)) &
                L('cached'))

    @property
    def _cond_http(self):
        return L('valid')

    def ev(self, item):
        return self.cond.ev(self.link)

    def describe(self, item, target):
        d = self.cond.describe(self.link, target)
        m = self.link.url
        return '{} -> {}'.format(m, d)


class ValidLinks(DynOr):

    @property
    def _sub_type(self):
        return ValidLink

    def _dyn_subs(self, item: ReleaseMonitor):
        return List.wrap(item.all_links)

    def describe(self, item, target):
        d = super().describe(item, target)
        return 'links => {}'.format(d)


class DownloaderProc(Process, Logging):

    def __init__(self, link, dir):
        self.link = link
        self.dir = dir
        self._downloader_fact = DownloaderFactory()
        self.done = Value('i', 0)
        self.fail = Value('i', 0)
        self.size = Value('i', -1)
        self.file_size = Value('i', -1)
        self.file_path = Array('c', 1024)
        self.error = Array('c', 1024)
        super().__init__(name='seriesd downloader')

    def run(self):
        self._download()
        self.done.value = 1

    def _download(self):
        try:
            dl = self._downloader_fact(self.link.download_url,
                                       download_dir=str(self.dir))
            dl.retrieve()
            self.size.value = dl.outfile_size
            self.file_size.value = dl.file_size
            self.file_path.value = str(dl.file_path).encode()
        except (NotEnoughDiskSpace, requests.RequestException,
                ShareHosterError, ConnectionError, socket.timeout) as e:
            self.error.value = str(e)
            self.fail.value = 1


class ProcAdapter:

    def __init__(self, proc: DownloaderProc, monitor) -> None:
        self.proc = proc
        self.monitor = monitor
        self.created = datetime.now()

    def kill(self):
        return self.proc.terminate()


@configurable(get=['download_dir', 'min_size', 'download_timeout'])
class Downloader(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'download manager')
        self._min_failed = 0
        self._proc = Empty()
        self._timeout = timedelta(minutes=self._download_timeout)

    def _fail(self, monitor, link, err):
        self.log.error('Downloader: {}'.format(err))
        f = monitor.failed_downloads + 1
        self._update(monitor, failed_downloads=f)
        self._releases.fail_link(link.id)

    def _handle(self, monitor):
        if self._proc.empty:
            self._proc = self._download(monitor)

    def _download(self, monitor):
        self.log.info('Downloading {!s}.'.format(monitor.release))
        self._update(monitor, downloading=True)
        link = monitor.link
        self._download_dir.mkdir(parents=True, exist_ok=True)
        proc = DownloaderProc(link, self._download_dir)
        proc.start()
        return Just(ProcAdapter(proc, monitor))

    def _cleanup(self):
        if (self._proc.exists(self._clean_done) or
                self._proc.exists(self._clean_timeouts)):
            self._proc = Empty()

    def _clean_done(self, a):
        proc = a.proc
        if proc.done.value:
            self._update_id(a.monitor.id, downloading=False)
            if proc.fail.value:
                self._fail(a.monitor, proc.link, proc.error.value)
            else:
                path = Path(proc.file_path.value.decode())
                return self._check_download(a.monitor, proc.link,
                                            proc.size.value,
                                            proc.file_size.value, path)

    def _clean_timeouts(self, a):
        if datetime.now() - a.created > self._timeout:
            a.kill()
            self._proc = Empty()

    def _check_download(self, monitor, link, size, file_size, file_path):
        ''' Verify the downloaded file's size by means of two criteria:
        The content-length header sent from the hoster must match
        exactly.
        The config parameter min_size defines what is considered to be a
        faulty file, i.e. the link checker didn't detect the file to be
        invalid, but it is small enough to just be the hoster's error
        page.
        '''
        if size < self._min_size:
            msg = 'File smaller than required ({}): {!s}'
            self._fail(monitor, link,
                       msg.format(sizeof_fmt(size), monitor.release))
        elif size != file_size and file_size != -1:
            msg = 'Filesize mismatch in download of {!s}:'
            self._fail(monitor, link, msg.format(monitor.release))
            msg = 'Got {}, expected {}.'
            self.log.error(msg.format(size, file_size))
            try:
                file_path.unlink()
            except IOError:
                msg = 'Couldn\'t delete broken download of {}.'
                self.log.warn(msg.format(monitor.release))
        else:
            self._releases.update_by_id(monitor.id,
                                        download_path=str(file_path),
                                        download_date_stamp=now_unix())
            self._releases.mark_episode_downloaded(monitor)
            return True

    @property
    def _conditions(self):
        return (~(R('downloaded') | R('nuked') | R('downloading')) &
                ValidLinks())

    def _choose(self, candidates):
        counts = [r.item.failed_downloads for r in candidates]
        min_failed = min(counts or [0])
        best = candidates.filter(
            lambda a: a.item.failed_downloads == min_failed)
        return best.head

    @property
    def _candidates(self):
        return self._releases.monitors.filter_by(downloaded=False).all()

__all__ = ('Downloader',)



import re
import subprocess
from dateutil.parser import parse as dateparse

from tek.errors import ParseError
from golgi.config import configurable

from amino import List, __, Maybe, _

from series.logging import Logging


@configurable(etvdb=['path'])
class ETVDB(Logging):

    def __call__(self, params, skip=1):
        args = [str(self._path)] + list(map(str, params))
        try:
            result = subprocess.run(args, timeout=60, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    universal_newlines=True)
        except subprocess.SubprocessError as e:
            self.log.error('etvdb subprocess error: {}'.format(e))
        else:
            if result.returncode == 0:
                output = List.lines(result.stdout)
                return List.wrap(output[skip:])
            else:
                cmd = ' '.join(args)
                output = List.lines(result.stderr)
                msg = 'etvdb returned {}:\n{}\n{}'
                self.log.error(msg.format(result.returncode, cmd, output))

    def single(self, params):
        result = self(params, 0)
        if isinstance(result, list):
            return result[0]
        elif isinstance(result, str):
            return result

    def m(self, params, skip=1):
        return Maybe(self(params, skip))


class Episode(object):

    def __init__(self, season, episode, title, overview, date):
        self.season = season
        self.episode = episode
        self.title = title
        self.overview = overview
        self.date = date

    @property
    def number(self):
        return self.episode

    def __str__(self):
        return '{}({}, {}, {})'.format(self.__class__.__name__, self.season,
                                       self.episode, self.date)

    @property
    def datetime(self):
        return dateparse(self.date)


class Show(object):

    def __init__(self, id, name, latest, next_):
        self.showid = id
        self.name = name
        self.latest = latest
        self.next_ = next_
        self.ended = False

    @property
    def latest_episode(self):
        return self.latest

    def __str__(self):
        return '{}({}, {}, {}, {})'.format(self.__class__.__name__,
                                           self.showid, self.name, self.latest,
                                           self.next_)


@configurable(etvdb=['series_name_map'])
class ETVDBFacade(object):

    def __init__(self):
        self._db = ETVDB()

    def season(self, name, season):
        params = self._season_params(name, season)
        output = self._db(params)
        if output:
            return [self._parse_episode(line) for line in output]
        else:
            return []

    def episode(self, name, season, episode):
        params = self._episode_params(name, season, episode)
        output = self._db(params)
        if output:
            return self._parse_episode(output[0])
        else:
            desc = '{} {}x{}'.format(name, season, episode)
            raise ParseError('Empty etvdb output for {}'.format(desc))

    def convert_date_enum(self, name, *date):
        name = self._translate_name(name)
        year, season, episode = date
        params = ['--name', name, '--date', '-'.join(date)]
        output = self._db(params)
        if output:
            data = self._parse_episode(output[0])
            season = data['season']
            episode = data['episode']
        return '{}x{}'.format(season, episode)

    def _season_params(self, name, season):
        return self._episode_params(name, season, 0)

    def _episode_params(self, name, season, episode):
        return ['-n', self._translate_name(name), '-s', str(season),
                '-e', str(episode)]

    def _parse_episode(self, line):
        parts = line.split('|')
        if len(parts) < 7:
            raise ParseError('Invalid etvdb output: {}'.format(line))
        return dict(
            episode=int(parts[1]),
            season=int(parts[0]),
            title=parts[3],
            overview=parts[5],
            date=parts[6]
        )

    def _translate_name(self, name):
        name = self._series_name_map.get(name, name)
        return re.sub('[. ]', '_', name)

    def id_by_name(self, name):
        return self._db.single(['-n', name, '-q', 'sid'])

    def ids_and_names_by_name(self, name):
        return self._db.m(['-f', name, '-q', 'sid'], skip=0)

    def query(self, id, attr):
        return self._db.single(['-N', str(id), '-q', attr])

    def episode_query(self, id, attr):
        output = self.query(id, attr)
        if output:
            data = self._parse_episode(output)
            return self.episode_model(data)

    def episode_model(self, data):
        return Episode(data['season'], data['episode'], data['title'],
                       data['overview'], data['date'])

    def show(self, name, sid=None):
        id = self.id_by_name(name) if sid is None else sid
        if id:
            sname = self.query(id, 'sname')
            latest = self.episode_query(id, 'aired_latest')
            nepi = self.episode_query(id, 'airs_next')
            return Show(id, sname, latest, nepi)

    def shows(self, name):
        return (
            (
                self.ids_and_names_by_name(name).eff() /
                __.split('|') /
                List.wrap
            ).value /
            _.unzip
        )

    def next_episode_date(self, show):
        if show.next_:
            return show.next_.datetime

    def next_episode_enum(self, show):
        if show.next_:
            return (show.next_.season, show.next_.episode)
        else:
            return [-1, -1]

    def airdate(self, dbshow, season, episode):
        show = self.show(dbshow.name, dbshow.etvdb_id)
        if show:
            data = self.episode(show.name, season, episode)
            if data:
                return self.episode_model(data).datetime

    def show_id_update_param(self, show):
        id = self.id_by_name(show.name)
        return dict(etvdb_id=id) if id else dict()

__all__ = ['ETVDBFacade']



import functools

from sqlpharmacy.core import Database

from sqlalchemy import Column, String


@functools.total_ordering
class Series(object, metaclass=Database.DefaultMeta):
    name = Column(String, nullable=False)
    subfps = Column(String)

    @property
    def formatted_name(self):
        return self.name.replace('_', ' ').title()

    @property
    def canonical_name(self):
        return self.name.replace(' ', '_').lower()

    def __str__(self):
        return self.canonical_name

    def __lt__(self, other):
        return isinstance(other, Series) and self.name < other.name

    def __eq__(self, other):
        return isinstance(other, Series) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    @property
    def info(self):
        return dict(
            id=self.id,
            name=self.formatted_name,
            new_count=self.new_count,
            subfps=self.subfps,
        )

    @property
    def new_count(self):
        return len([e for e in self.episodes if e.new])

    @property
    def empty(self):
        return all(s.empty for s in self.seasons) or self.seasons.count == 0

__all__ = ['Series']



from mpv import MPV as Player, _mpv_terminate_destroy

from series.library.player.interface import PlayerInterface


class Callback:

    def __init__(self, func):
        self._func = func

    def call(self):
        self._func()


class MPV(PlayerInterface):

    def __init__(self, args):
        super().__init__(Player(**args))
        self._player.event_callbacks.append(Callback(self.event))

    def stop(self):
        _mpv_terminate_destroy(self._player.handle)
        self._player = None

    def pause(self):
        self._player.pause = not self._player.pause.val

    def osd(self, level):
        self._player.osd_level = level

    def osd_level(self):
        return self._player.osd_level

    @property
    def current_volume(self):
        return self._player.ao_volume

    def volume(self, value):
        self._player.ao_volume = value

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except TypeError:
            pass

    def event(self):
        try:
            pass
        except Exception as e:
            self.log.error(e)

__all__ = ['MPV']



from series.condition import AttrCondition
from series.handler import Handler
from series.get.model.release import ReleaseMonitor
from series.get.model.link import Link, Torrent
from series.get.model.show import Show

from sqlalchemy.sql.expression import and_


class ReleaseAttr(AttrCondition[ReleaseMonitor]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('release', attr, target=target)

R = ReleaseAttr


class LinkAttr(AttrCondition[Link]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('link', attr, target)


L = LinkAttr


class ShowAttr(AttrCondition[Show]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('show', attr, target)


S = ShowAttr


class BaseHandler(Handler):

    def __init__(self, data, interval, description, **kw):
        self._data = data
        super().__init__(interval, description, **kw)

    @property
    def _candidates(self):
        return self._data.all

    @property
    def _lock(self):
        return self._data.lock

    def _update(self, item, **data):
        self._update_id(item.id, **data)

    def _update_id(self, id, **data):
        self._data.update_by_id(id, **data)


class ReleaseHandler(BaseHandler):

    @property
    def _releases(self):
        return self._data

    @property
    def _no_cached_torrents_q(self):
        return (
            self._releases.monitors
            .filter_by(downloaded=False, downloading=False)
            .filter(~ReleaseMonitor.torrents.any(
                and_(Torrent.cached, ~Torrent.dead)))
        )

    @property
    def _no_cached_torrents(self):
        return self._no_cached_torrents_q.all()


class ShowHandler(BaseHandler):

    @property
    def _shows(self):
        return self._data

__all__ = ['ReleaseHandler', 'ShowHandler']
